@@ -57,6 +57,9 @@ valid across interpreter invocations, unlike Python's built-in hashes.
 This module also provides a disk-backed dictionary that uses persistent hashing.
 
 .. autoexception:: NoSuchEntryError
+.. autoexception:: NoSuchEntryInvalidKeyError
+.. autoexception:: NoSuchEntryInvalidContentsError
+.. autoexception:: NoSuchEntryCollisionError
 .. autoexception:: ReadOnlyEntryError
 
 .. autoexception:: CollisionWarning
@@ -67,22 +70,6 @@ This module also provides a disk-backed dictionary that uses persistent hashing.
 """
 
 
-def _make_dir_recursively(dir_):
-    try:
-        os.makedirs(dir_)
-    except OSError as ex:
-        from errno import EEXIST
-        if ex.errno != EEXIST:
-            raise
-
-
-def update_checksum(checksum, obj):
-    if isinstance(obj, str):
-        checksum.update(obj.encode("utf8"))
-    else:
-        checksum.update(obj)
-
-
 # {{{ cleanup managers
 
 class CleanupBase:
@@ -221,7 +208,7 @@ class KeyBuilder:
 
         digest = getattr(key, "_pytools_persistent_hash_digest", None)
 
-        if digest is None:
+        if digest is None and not isinstance(key, type):
             try:
                 method = key.update_persistent_hash
             except AttributeError:
@@ -238,26 +225,35 @@ class KeyBuilder:
             try:
                 method = getattr(self, "update_for_"+tname)
             except AttributeError:
-                if (
-                        # Handling numpy >= 1.20, for which
-                        # type(np.dtype("float32")) -> "dtype[float32]"
-                        tname.startswith("dtype[")
-                        # Handling numpy >= 1.25, for which
-                        # type(np.dtype("float32")) -> "Float32DType"
-                        or tname.endswith("DType")
-                        ) and "numpy" in sys.modules:
+                if "numpy" in sys.modules:
                     import numpy as np
-                    if isinstance(key, np.dtype):
-                        method = self.update_for_specific_dtype
 
-                elif issubclass(tp, Enum):
-                    method = self.update_for_enum
+                    # Hashing numpy dtypes
+                    if (
+                            # Handling numpy >= 1.20, for which
+                            # type(np.dtype("float32")) -> "dtype[float32]"
+                            tname.startswith("dtype[")
+                            # Handling numpy >= 1.25, for which
+                            # type(np.dtype("float32")) -> "Float32DType"
+                            or tname.endswith("DType")
+                            ):
+                        if isinstance(key, np.dtype):
+                            method = self.update_for_specific_dtype
 
-                elif is_dataclass(tp):
-                    method = self.update_for_dataclass
+                    # Hashing numpy scalars
+                    elif isinstance(key, np.number):
+                        # Non-numpy scalars are handled above in the try block.
+                        method = self.update_for_numpy_scalar
 
-                elif _HAS_ATTRS and attrs.has(tp):
-                    method = self.update_for_attrs
+                if method is None:
+                    if issubclass(tp, Enum):
+                        method = self.update_for_enum
+
+                    elif is_dataclass(tp):
+                        method = self.update_for_dataclass
+
+                    elif _HAS_ATTRS and attrs.has(tp):
+                        method = self.update_for_attrs
 
             if method is not None:
                 inner_key_hash = self.new_hash()
@@ -292,6 +288,8 @@ class KeyBuilder:
         key_hash.update(
             f"{key.__module__}.{key.__qualname__}.{key.__name__}".encode("utf-8"))
 
+    update_for_ABCMeta = update_for_type  # noqa: N815
+
     @staticmethod
     def update_for_int(key_hash, key):
         sz = 8
@@ -314,6 +312,10 @@ class KeyBuilder:
     def update_for_float(key_hash, key):
         key_hash.update(key.hex().encode("utf8"))
 
+    @staticmethod
+    def update_for_complex(key_hash, key):
+        key_hash.update(repr(key).encode("utf-8"))
+
     @staticmethod
     def update_for_str(key_hash, key):
         key_hash.update(key.encode("utf8"))
@@ -333,8 +335,10 @@ class KeyBuilder:
             key_hash,
             (self.rec(self.new_hash(), key_i).digest() for key_i in key))
 
+    update_for_FrozenOrderedSet = update_for_frozenset  # noqa: N815
+
     @staticmethod
-    def update_for_NoneType(key_hash, key):  # noqa
+    def update_for_NoneType(key_hash, key):  # noqa: N802
         del key
         key_hash.update(b"<None>")
 
@@ -350,20 +354,42 @@ class KeyBuilder:
     def update_for_specific_dtype(key_hash, key):
         key_hash.update(key.str.encode("utf8"))
 
+    @staticmethod
+    def update_for_numpy_scalar(key_hash, key):
+        import numpy as np
+        if hasattr(np, "complex256") and key.dtype == np.dtype("complex256"):
+            key_hash.update(repr(complex(key)).encode("utf8"))
+        elif hasattr(np, "float128") and key.dtype == np.dtype("float128"):
+            key_hash.update(repr(float(key)).encode("utf8"))
+        else:
+            key_hash.update(np.array(key).tobytes())
+
     def update_for_dataclass(self, key_hash, key):
-        self.rec(key_hash, type(key_hash).__name__.encode("utf-8"))
+        self.rec(key_hash, f"{type(key).__qualname__}.{type(key).__name__}")
 
         for fld in dc_fields(key):
             self.rec(key_hash, fld.name)
             self.rec(key_hash, getattr(key, fld.name, None))
 
     def update_for_attrs(self, key_hash, key):
-        self.rec(key_hash, type(key_hash).__name__.encode("utf-8"))
+        self.rec(key_hash, f"{type(key).__qualname__}.{type(key).__name__}")
 
         for fld in attrs.fields(key.__class__):
             self.rec(key_hash, fld.name)
             self.rec(key_hash, getattr(key, fld.name, None))
 
+    def update_for_frozendict(self, key_hash, key):
+        from pytools import unordered_hash
+
+        unordered_hash(
+            key_hash,
+            (self.rec(self.new_hash(), (k, v)).digest() for k, v in key.items()))
+
+    update_for_immutabledict = update_for_frozendict
+    update_for_constantdict = update_for_frozendict
+    update_for_PMap = update_for_frozendict  # noqa: N815
+    update_for_Map = update_for_frozendict  # noqa: N815
+
     # }}}
 
 # }}}
@@ -494,6 +520,18 @@ class NoSuchEntryError(KeyError):
     pass
 
 
+class NoSuchEntryInvalidKeyError(NoSuchEntryError):
+    pass
+
+
+class NoSuchEntryInvalidContentsError(NoSuchEntryError):
+    pass
+
+
+class NoSuchEntryCollisionError(NoSuchEntryError):
+    pass
+
+
 class ReadOnlyEntryError(KeyError):
     pass
 
@@ -518,8 +556,15 @@ class _PersistentDictBase:
             except ImportError:
                 import appdirs
 
+            if sys.platform == "darwin" and os.getenv("XDG_CACHE_HOME") is not None:
+                # appdirs and platformdirs do not handle XDG_CACHE_HOME on macOS
+                # https://github.com/platformdirs/platformdirs/issues/269
+                cache_dir = join(os.getenv("XDG_CACHE_HOME"), "pytools")
+            else:
+                cache_dir = appdirs.user_cache_dir("pytools", "pytools")
+
             container_dir = join(
-                    appdirs.user_cache_dir("pytools", "pytools"),
+                    cache_dir,
                     "pdict-v4-{}-py{}".format(
                         identifier,
                         ".".join(str(i) for i in sys.version_info)))
@@ -579,7 +624,7 @@ class _PersistentDictBase:
         return join(self.container_dir, str(hexdigest_key) + ".lock")
 
     def _make_container_dir(self):
-        _make_dir_recursively(self.container_dir)
+        os.makedirs(self.container_dir, exist_ok=True)
 
     def _collision_check(self, key, stored_key, _stacklevel):
         if stored_key != key:
@@ -595,7 +640,7 @@ class _PersistentDictBase:
             # This is here so we can step through equality comparison to
             # see what is actually non-equal.
             stored_key == key  # pylint:disable=pointless-statement  # noqa: B015
-            raise NoSuchEntryError(key)
+            raise NoSuchEntryCollisionError(key)
 
     def __getitem__(self, key):
         return self.fetch(key, _stacklevel=1)
@@ -623,6 +668,7 @@ class WriteOncePersistentDict(_PersistentDictBase):
     .. automethod:: __getitem__
     .. automethod:: __setitem__
     .. automethod:: clear
+    .. automethod:: clear_in_mem_cache
     .. automethod:: store
     .. automethod:: store_if_not_present
     .. automethod:: fetch
@@ -637,7 +683,15 @@ class WriteOncePersistentDict(_PersistentDictBase):
             *in_mem_cache_size* items
         """
         _PersistentDictBase.__init__(self, identifier, key_builder, container_dir)
-        self._cache = _LRUCache(in_mem_cache_size)
+        self._in_mem_cache_size = in_mem_cache_size
+        self.clear_in_mem_cache()
+
+    def clear_in_mem_cache(self) -> None:
+        """
+        .. versionadded:: 2023.1.1
+        """
+
+        self._cache = _LRUCache(self._in_mem_cache_size)
 
     def _spin_until_removed(self, lock_file, stacklevel):
         from os.path import exists
@@ -740,7 +794,7 @@ class WriteOncePersistentDict(_PersistentDictBase):
                     f"Remove the directory '{item_dir}' if necessary. "
                     f"(caught: {type(e).__name__}: {e})",
                     stacklevel=1 + _stacklevel)
-            raise NoSuchEntryError(key)
+            raise NoSuchEntryInvalidKeyError(key)
 
         self._collision_check(key, read_key, 1 + _stacklevel)
 
@@ -753,12 +807,13 @@ class WriteOncePersistentDict(_PersistentDictBase):
 
         try:
             read_contents = self._read(contents_file)
-        except Exception:
+        except Exception as e:
             self._warn(f"{type(self).__name__}({self.identifier}) "
-                    f"encountered an invalid key file for key {hexdigest_key}. "
-                    f"Remove the directory '{item_dir}' if necessary.",
+                    f"encountered an invalid contents file for key {hexdigest_key}. "
+                    f"Remove the directory '{item_dir}' if necessary."
+                    f"(caught: {type(e).__name__}: {e})",
                     stacklevel=1 + _stacklevel)
-            raise NoSuchEntryError(key)
+            raise NoSuchEntryInvalidContentsError(key)
 
         # }}}
 
@@ -849,13 +904,14 @@ class PersistentDict(_PersistentDictBase):
 
                 try:
                     read_key = self._read(key_path)
-                except Exception:
+                except Exception as e:
                     item_dir_m.reset()
                     self._warn(f"{type(self).__name__}({self.identifier}) "
                             "encountered an invalid key file for key "
-                            f"{hexdigest_key}. Entry deleted.",
+                            f"{hexdigest_key}. Entry deleted."
+                            f"(caught: {type(e).__name__}: {e})",
                             stacklevel=1 + _stacklevel)
-                    raise NoSuchEntryError(key)
+                    raise NoSuchEntryInvalidKeyError(key)
 
                 self._collision_check(key, read_key, 1 + _stacklevel)
 
@@ -868,13 +924,14 @@ class PersistentDict(_PersistentDictBase):
 
                 try:
                     read_contents = self._read(value_path)
-                except Exception:
+                except Exception as e:
                     item_dir_m.reset()
                     self._warn(f"{type(self).__name__}({self.identifier}) "
-                            "encountered an invalid key file for key "
-                            f"{hexdigest_key}. Entry deleted.",
+                            "encountered an invalid contents file for key "
+                            f"{hexdigest_key}. Entry deleted."
+                            f"(caught: {type(e).__name__}: {e})",
                             stacklevel=1 + _stacklevel)
-                    raise NoSuchEntryError(key)
+                    raise NoSuchEntryInvalidContentsError(key)
 
                 return read_contents
 
@@ -907,13 +964,14 @@ class PersistentDict(_PersistentDictBase):
 
                 try:
                     read_key = self._read(key_file)
-                except Exception:
+                except Exception as e:
                     item_dir_m.reset()
                     self._warn(f"{type(self).__name__}({self.identifier}) "
                             "encountered an invalid key file for key "
-                            f"{hexdigest_key}. Entry deleted",
+                            f"{hexdigest_key}. Entry deleted"
+                            f"(caught: {type(e).__name__}: {e})",
                             stacklevel=1 + _stacklevel)
-                    raise NoSuchEntryError(key)
+                    raise NoSuchEntryInvalidKeyError(key)
 
                 self._collision_check(key, read_key, 1 + _stacklevel)
 
